%include "x86_64/def.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;header

%define HEAP_SIZE (heap_end-heap_beg)
%define EMPTY     (empty-heap_beg)
%define UNDEF     (undef-heap_beg)
%define BOOL_F    (bool_f-heap_beg)
%define BOOL_T    (bool_t-heap_beg)
%define START     (start-heap_beg)

section header
header:
  dq      0         ;B
  dq      HEAP_SIZE ;F
  dq      HEAP_SIZE ;L
  dq      EMPTY     ;S
  dq      EMPTY     ;E
  dq      START     ;C
  dq      EMPTY     ;D

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;macros for constructing lists
;;C_* = cons with cdr = next (cons addr + 1)
;;L_* = cons with cdr = ()

;car/cdr or vector elt
%macro REF 1 ;addr
  dq      %1
%endmacro

;cons cell
%macro C 2 ;car cdr
  REF     %1
  REF     %2
%endmacro

;vector
%macro V 1 ;size
  dq      (TYPE_VECTOR<<TYPE_SHIFT) | %1
%endmacro

;integer
%macro INT 1 ;int_val
  dq      TYPE_FIXINT<<TYPE_SHIFT
  dq      %1
%endmacro

;char
%macro CHAR 1 ;char_code
  dq      (TYPE_CHAR<<TYPE_SHIFT) | %1
%endmacro

;3 char symbol
%macro SYM 3 ;char_code*3
  ;i would like to do this but im not allowed by piss of shit asembler
;  dq      (TYPE_SYMBOL<<TYPE_SHIFT) | %%vec
  dd      %%vec
  dd      TYPE_SYMBOL<<(TYPE_SHIFT-32)
%%vec:
  V       3
  REF     %%1
  REF     %%2
  REF     %%3
%%1:
  CHAR    %1
%%2:
  CHAR    %2
%%3:
  CHAR    %3
%endmacro

;cons with car = any value, cdr = next (cons addr + 1)
%macro C_ADDR 1 ;car
  C       %1, %%next
%%next:
%endmacro

;cons with car = int, cdr = next
%macro C_INT 1 ;int_val
  C       %%int, %%next
%%int:
  INT     %1
%%next:
%endmacro

;cons with car = int, cdr = ()
%macro L_INT 1 ;int_val
  C       %%int, EMPTY
%%int:
  INT     %1
%endmacro

;cons with car = char, cdr = next
%macro C_CHAR 1 ;char_code
  C       %%char, %%next
%%char:
  CHAR    %1
%%next:
%endmacro

;cons with car = sym, cdr = next
%macro C_SYM 3 ;char_code*3
  C       %%sym, %%next
%%sym:
  SYM     %1, %2, %3
%%next:
%endmacro

;cons with car = sym, cdr = ()
%macro L_SYM 3 ;char_code*3
  C       %%sym, EMPTY
%%sym:
  SYM     %1, %2, %3
%%next:
%endmacro

;cons with car = (x . y) where xy are ints, cdr = next
%macro C_IDX 2 ;x y
  C       %%idx, %%next
%%idx:
  C       %%x, %%y
%%x:
  INT     %1
%%y:
  INT     %2
%%next:
%endmacro

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;macros for writing code

%macro LDC 1 ;addr
  C_INT   OP_LDC
  C_ADDR  %1
%endmacro

%macro LDCI 1 ;int_val
  C_INT   OP_LDC
  C_INT   %1
%endmacro

%macro LDCC 1 ;char_code
  C_INT   OP_LDC
  C_CHAR  %1
%endmacro

%macro LD 2 ;x y
  C_INT   OP_LD
  C_IDX   %1, %2
%endmacro

%macro ST 2 ;x y
  C_INT   OP_ST
  C_IDX   %1, %2
%endmacro

%macro LDF 1 ;code_addr
  C_INT   OP_LDF
  C_ADDR  %1
%endmacro

%macro SEL 2 ;true_addr false_addr
  C_INT   OP_SEL
  C_ADDR  %1
  C_ADDR  %2
%endmacro

%macro TEST 1 ;true_addr
  C_INT   OP_TEST
  C_ADDR  %1
%endmacro

%define NIL     C_INT OP_NIL
%define POP     C_INT OP_POP
%define TYPE    C_INT OP_TYPE
%define V2S     C_INT OP_V2S
%define V2T     C_INT OP_V2T
%define S2V     C_INT OP_S2V
%define I2C     C_INT OP_I2C
%define C2I     C_INT OP_C2I
%define NULL    C_INT OP_NULL
%define CAR     C_INT OP_CAR
%define CDR     C_INT OP_CDR
%define RPLACA  C_INT OP_RPLACA
%define CONS    C_INT OP_CONS
%define VEC     C_INT OP_VEC
%define VREF    C_INT OP_VREF
%define VSET    C_INT OP_VSET
%define ADD     C_INT OP_ADD
%define SUB     C_INT OP_SUB
%define MUL     C_INT OP_MUL
%define EQ      C_INT OP_EQ
%define EQI     C_INT OP_EQI
%define LT      C_INT OP_LT
%define JOIN    L_INT OP_JOIN
%define AP      C_INT OP_AP
%define RTN     L_INT OP_RTN
%define RECV    C_INT OP_RECV
%define IDEF    C_INT OP_IDEF
%define READC   C_INT OP_READC
%define IMG     L_INT OP_IMG

;call by ldf, excludes arg list
%macro CALL 1
  LDF     %1
  AP
  RECV
%endmacro

;call with 0 args
%macro CALL0 1
  NIL
  CALL    %1
%endmacro

%define LET CALL

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;heap

section heap vstart=0
heap_beg:
empty:
  dq      TYPE_EMPTY<<TYPE_SHIFT ;()
undef:
  dq      TYPE_UNDEF<<TYPE_SHIFT ;undefined
bool_f:
  dq      TYPE_BOOL<<TYPE_SHIFT ;#f
bool_t:
  dq      TYPE_BOOL<<TYPE_SHIFT ;#t

;-------------------------------------------------------------------------------
;reader

;3char vector equals?
chvec_equals: ;vec0 vec1 idx
  LD      0, 2 ;idx
  LDCI    3
  EQI
  TEST    .done
  LD      0, 2 ;idx
  LD      0, 0 ;vec0
  VREF
  C2I
  LD      0, 2 ;idx
  LD      0, 1 ;vec1
  VREF
  C2I
  EQI
  TEST    .char_equal
  LDC     BOOL_F
  RTN
.char_equal:
  ;recurse
  NIL
    LD      0, 2
    LDCI    1
    ADD           ;idx
  CONS
    LD      0, 1  ;vec1
  CONS
    LD      0, 0  ;vec2
  CONS
  CALL    chvec_equals
  RTN
.done:
  ;we done, vecs are equal
  LDC     BOOL_T
  RTN

;symbol lookup, #f if not found
lookup: ;vec tab
  LD      0, 1
  NULL
  TEST    .done
  NIL
    LDCI    0     ;idx
  CONS
    LD      0, 1
    CAR
    S2V           ;vec1
  CONS
    LD      0, 0  ;vec0
  CONS
  CALL    chvec_equals
  TEST    .equal
  ;recurse
  NIL
    LD      0, 1
    CDR           ;tab
  CONS
    LD      0, 0  ;vec
  CONS
  CALL    lookup
  RTN
.equal:
  ;char vecs were equal, return the symbol
  LD      0, 1
  CAR
  RTN
.done:
  ;symbol not exist in table
  LDC     BOOL_F
  RTN

;intern symbol, return sym
intern: ;sym
  NIL
    NIL
      LDC     sym_tab
      CAR               ;tab
    CONS
      LD      0, 0
      S2V               ;vec
    CONS
    CALL    lookup    ;lookup_sym
  CONS
  LET     .let
  RTN
.let:
  LD      0, 0 ;lookup_sym
  TEST    .exists
  ;didnt exist, cons sym to table and return
  LDC     sym_tab
  CAR
  LD      1, 0 ;sym
  CONS
  LDC     sym_tab
  RPLACA
  POP
  LD      1, 0 ;sym
  RTN
.exists:
  ;existed, return with no table modifications
  LD      0, 0 ;sym
  RTN

;read char
read_char:
  IDEF
  READC
  RTN

;read char code (integer)
read_charcode:
  CALL0   read_char
  C2I
  RTN

;skip to end of line, return lf
skip_line:
  NIL
    CALL0   read_charcode ;char
  CONS
  LET     .let
  RTN
.let:
  LD      0, 0
  LDCI    10 ;ascii lf
  EQI
  TEST    .done
  CALL0   skip_line
  RTN
.done:
  LD      0, 0
  RTN

;skip whitespace+comments, return first non-ws char
skip:
  NIL
    CALL0   read_charcode ;char
  CONS
  LET     .let
  RTN
.let:
  LD      0, 0
  LDCI    ' '
  EQI
  TEST    skip
  LD      0, 0
  LDCI    10 ;ascii lf
  EQI
  TEST    skip
  LD      0, 0
  LDCI    ';'
  EQI
  SEL     .semi, .done
  RTN
.semi:
  CALL0   skip_line
  POP
  CALL0   skip
  JOIN
.done:
  ;not whitespace or comment, return char
  LD      0, 0
  JOIN

;read and return list
read_list:
  NIL
    CALL0   skip ;char
  CONS
  LET     .let
  RTN
.let:
  LD      0, 0 ;char
  LDCI    ')'
  EQI
  TEST    .rightp
  NIL
    NIL
      LD      0, 0  ;initial_char
    CONS
    CALL    read  ;datum
  CONS
  LET     .let2
  RTN
.let2:
  CALL0   read_list
  LD      0, 0 ;datum
  CONS
  RTN
.rightp
  ;end of list reached, ret ()
  NIL
  RTN

;read and return expression
read: ;initial_char
  LD      0, 0
  LDCI    '('
  EQI
  TEST    .leftp
  LDCI    'A'
  LD      0, 0
  LT      ;if < A, interpret as integer. otherwise, symbol
  TEST    .integer
  ;read 3char symbol and intern
  NIL
    LD      0, 0
    I2C
    LDCI    3
    VEC
  CONS
  LET     .let
  RTN
.let:
  NIL
    CALL0   read_char
    LDCI    1
    LD      0, 0
    VSET
    POP
    CALL0   read_char
    LDCI    2
    LD      0, 0
    VSET
    POP
    LD      0, 0
    V2S           ;sym
  CONS
  CALL    intern
  RTN
.integer:
  ;read 3digit integer
  LDCI    '0'
  LD      0, 0
  SUB     ;get numeric value of digit
  LDCI    100
  MUL     ;mul by its position in the integer
  LDCI    '0'
  CALL0   read_charcode
  SUB
  LDCI    10
  MUL
  LDCI    '0'
  CALL0   read_charcode
  SUB
  ADD
  ADD
  RTN
.leftp:
  CALL0   read_list
  RTN

;-------------------------------------------------------------------------------
;compiler

;position of symbol in list, #f if not exist
position: ;sym lst idx
  LD      0, 1
  NULL
  TEST    .null
  LD      0, 1
  CAR
  LD      0, 0
  EQ
  TEST    .eq
  ;recurse
  NIL
    LD      0, 2
    LDCI    1
    ADD           ;idx
  CONS
    LD      0, 1
    CDR           ;lst
  CONS
    LD      0, 0  ;sym
  CONS
  CALL    position
  RTN
.eq:
  ;found, return index
  LD      0, 2
  RTN
.null:
  ;not found, return #f
  LDC     BOOL_F
  RTN

;get (x . y) index of symbol from lexical environment, assumes exists
index: ;sym env x
  NIL
    NIL
      LDCI    0         ;idx
    CONS
      LD      0, 1
      CAR               ;lst
    CONS
      LD      0, 0      ;sym
    CONS
    CALL    position  ;y
  CONS
  LET     .let
  RTN
.let:
  LD      0, 0 ;y
  TEST    .found
  ;recurse
  NIL
    LD      1, 2
    LDCI    1
    ADD           ;x
  CONS
    LD      1, 1
    CDR           ;env
  CONS
    LD      1, 0  ;sym
  CONS
  CALL    index
  RTN
.found:
  ;found, return (x . y)
  LD      0, 0 ;y
  LD      1, 2 ;x
  CONS
  RTN

;compile value list (not includes initial NIL)
compile_value_list: ;arg env acc
  LD      0, 0
  NULL
  TEST    .null
  ;cons compiled arg onto vl and recurse
  NIL
    NIL
      LD      0, 2
      LDCI    OP_CONS
      CONS            ;acc
    CONS
      LD      0, 1    ;env
    CONS
      LD      0, 0
      CAR             ;arg
    CONS
    CALL    compile ;acc
  CONS
    LD      0, 1    ;env
  CONS
    LD      0, 0
    CDR             ;arg
  CONS
  CALL    compile_value_list
  RTN
.null: ;done, return acc
  LD      0, 2
  RTN

;compile empty value list (not includes initial NIL)
compile_empty_list: ;binds acc
  LD      0, 0
  NULL
  TEST    .null
  ;cons () onto vl and recurse
  NIL
    LD      0, 1  ;acc
  CONS
    LD      0, 0
    CDR           ;binds
  CONS
  CALL    compile_empty_list
  LDCI    OP_CONS
  CONS
  LDCI    OP_NIL
  CONS
  RTN
.null: ;done, return acc
  LD      0, 1
  RTN

;compile pushing of built in operator args
compile_op_args: ;arg env acc
  LD      0, 0
  NULL
  TEST    .null
  ;compile first arg and recurse
  NIL
    NIL
      LD      0, 2    ;acc
    CONS
      LD      0, 1    ;env
    CONS
      LD      0, 0
      CAR             ;expr
    CONS
    CALL    compile ;acc
  CONS
    LD      0, 1    ;env
  CONS
    LD      0, 0
    CDR             ;arg
  CONS
  CALL    compile_op_args
  RTN
.null: ;done, return acc
  LD      0, 2
  RTN

;compile built in operator
compile_opp: ;arg env acc
  NIL
    LD      0, 2
    LD      0, 0
    CAR
    CONS          ;acc
  CONS
    LD      0, 1  ;env
  CONS
    LD      0, 0
    CDR           ;arg
  CONS
  CALL    compile_op_args
  RTN

;compile if expresion
;arg0 = cond, arg1 = then, arg2 = else
compile_iff: ;arg env acc
  NIL
    LD      0, 2

    NIL
      NIL
      LDCI    OP_JOIN
      CONS            ;acc
    CONS
      LD      0, 1    ;env
    CONS
      LD      0, 0
      CDR
      CDR
      CAR             ;expr
    CONS
    CALL    compile   ;compile else
    CONS

    NIL
      NIL
      LDCI    OP_JOIN
      CONS            ;acc
    CONS
      LD      0, 1    ;env
    CONS
      LD      0, 0
      CDR
      CAR             ;expr
    CONS
    CALL    compile   ;compile then
    CONS

    LDCI    OP_SEL
    CONS            ;acc
  CONS
    LD      0, 1    ;env
  CONS
    LD      0, 0
    CAR             ;expr
  CONS
  CALL    compile   ;compile cond
  RTN

;get list of binding symbols ((a 0) (b 1)) => (a b)
map_car_binds: ;binds
  LD      0, 0
  NULL
  TEST    .null
  NIL
    LD      0, 0
    CDR           ;binds
  CONS
  CALL    map_car_binds
  LD      0, 0
  CAR
  CAR
  CONS
  RTN
.null:
  NIL
  RTN

;compile setting of vals for letrec
;([<val_y> ST (0 . y) POP]...)
compile_setvals: ;binds env acc
  LD      0, 0
  NULL
  TEST    .null
  NIL
    NIL
      LD      0, 2  ;acc
    CONS
      LD      0, 1  ;env
    CONS
      LD      0, 0
      CDR           ;binds
    CONS
    CALL    compile_setvals ;recurse
    LDCI    OP_POP
    CONS
    NIL
      LDCI    0     ;x
    CONS
      LD      0, 1  ;env
    CONS
      LD      0, 0
      CAR
      CAR           ;sym
    CONS
    CALL    index
    CONS
    LDCI    OP_ST
    CONS          ;acc
  CONS
    LD      0, 1  ;env
  CONS
    LD      0, 0
    CAR
    CDR
    CAR           ;expr
  CONS
  CALL    compile ;compile value
  RTN
.null: ;done, return acc
  LD      0, 2
  RTN

;compile letrec
compile_let: ;arg env acc
  NIL
    LD      0, 1
    NIL
      LD      0, 0
      CAR             ;binds ((a 0) (b 1)...)
    CONS
    CALL    map_car_binds
    CONS            ;new_env
  CONS
  LET     .let
  RTN
.let:
  NIL
    LD      1, 2
    LDCI    OP_RECV
    CONS
    LDCI    OP_AP
    CONS
    NIL
      NIL
        NIL
        LDCI    OP_RTN
        CONS            ;acc
      CONS
        LD      0, 0    ;env (new_env)
      CONS
        LD      1, 0
        CDR
        CAR             ;expr
      CONS
      CALL    compile ;acc
    CONS
      LD      0, 0    ;env (new_env)
    CONS
      LD      1, 0
      CAR             ;binds
    CONS
    CALL    compile_setvals
    CONS
    LDCI    OP_LDF
    CONS            ;acc
  CONS
    LD      1, 0
    CAR             ;binds
  CONS
  CALL    compile_empty_list
  LDCI    OP_NIL
  CONS
  RTN

;compile lambada
compile_lam: ;arg env acc
  LD      0, 2 ;acc
  NIL
    NIL
    LDCI    OP_RTN
    CONS            ;acc
  CONS
    LD      0, 1
    LD      0, 0
    CAR
    CONS            ;env (new_env)
  CONS
    LD      0, 0
    CDR
    CAR             ;expr
  CONS
  CALL    compile
  CONS
  LDCI    OP_LDF
  CONS
  RTN

;compile expression
;expr = expression to be compiled
;env  = list of lists of symbols in environment levels
;acc  = tail of generated code
compile: ;expr env acc
  NIL
    LD      0, 0
    TYPE          ;expr_type
  CONS
  LET     .let
  RTN
.let:
  ;figure type
  LD      0, 0
  LDCI    TYPE_FIXINT
  EQI
  TEST    .fixint
  LD      0, 0
  LDCI    TYPE_SYMBOL
  EQI
  TEST    .symbol

.cons: ;assume type is cons (syntax or call)
  NIL
    NIL
      LDCI    0         ;idx
    CONS
      LDC     keywords  ;lst
    CONS
      LD      1, 0
      CAR               ;sym
    CONS
    CALL    position  ;syntax_idx
  CONS
    LD      1, 0
    CDR               ;arg
  CONS
    LD      1, 0
    CAR               ;func
  CONS
  LET     .let2
  RTN
.let2:
  LD      0, 2  ;syntax_idx
  TEST    .syntax

.call: ;(NIL [<arg> CONS]... <func> AP RECV)
  NIL             ;for compile_value_list
    NIL               ;for compile
      LD      2, 2
      LDCI    OP_RECV
      CONS
      LDCI    OP_AP
      CONS            ;acc
    CONS
      LD      2, 1    ;env
    CONS
      LD      0, 0    ;func
    CONS
    CALL    compile ;acc
  CONS
    LD      2, 1    ;env
  CONS
    LD      0, 1    ;arg
  CONS
  CALL    compile_value_list
  LDCI    OP_NIL
  CONS
  RTN

.syntax: ;behavior depends on keyword
  NIL
    LD      2, 2  ;acc
  CONS
    LD      2, 1  ;env
  CONS
    LD      0, 1  ;arg
  CONS
  ;load syntax func from vector and call
  LD      0, 2  ;syntax_idx
  LDC     keyfunc
  VREF
  AP
  RECV
  RTN

.symbol: ;(LD (x . y))
  LD      1, 2  ;acc
  NIL
    LDCI    0     ;x
  CONS
    LD      1, 1  ;env
  CONS
    LD      1, 0  ;sym
  CONS
  CALL    index
  CONS
  LDCI    OP_LD
  CONS
  RTN

.fixint: ;(LDC expr)
  LD      1, 2 ;acc
  LD      1, 0 ;expr
  CONS
  LDCI    OP_LDC
  CONS
  RTN

;-------------------------------------------------------------------------------

;symbol table (cons so its setable)
sym_tab:
  C       keywords, EMPTY ;initial state of sym tab

;list of syntactic keywords. also used as initial symbol table
keywords:
  C_SYM   'o', 'p', 'p'
  C_SYM   'i', 'f', 'f'
  C_SYM   'l', 'e', 't'
  L_SYM   'l', 'a', 'm'

;vector of compiler bindings (index 0 = opp, ...)
keyfunc:
  V       4
  REF     EMPTY
  REF     EMPTY
  REF     EMPTY
  REF     EMPTY

;path to dumped image
image_path:
  V       3
  REF     .char0
  REF     .char1
  REF     .char2
.char0:
  CHAR    'T'
.char1:
  CHAR    'L'
.char2:
  CHAR    '1'

;-------------------------------------------------------------------------------
;entry point

%macro COMPILER_BIND 2
  LDF     %2
  LDCI    %1
  LDC     keyfunc
  VSET
  POP
%endmacro

start:
  COMPILER_BIND 0, compile_opp
  COMPILER_BIND 1, compile_iff
  COMPILER_BIND 2, compile_let
  COMPILER_BIND 3, compile_lam

  ;img arg1 - compiled code
  NIL             ;for compile
    NIL
    LDCI    OP_STOP
    CONS
    LDC     BOOL_T
    CONS
    LDCI    OP_LDC
    CONS            ;acc
  CONS
    NIL             ;env
  CONS
    NIL
      CALL0   skip    ;char
    CONS
    CALL    read    ;expr
  CONS
  CALL    compile

  ;img arg0 - output image path
  LDC     image_path
  V2T     ;turn into string
  IMG

heap_end:

