%define HEADER_SIZE (7*8) ;7 64bit regs (H+N excluded)

section .text

open_image:
  mov     rdx, FILE_MODE ;mode
  call    open
  cmp     rax, 0
  jl      eggog_open
  ret

read_image:
  call    read
  cmp     rax, rdx
  jne     eggog_read
  ret

write_image:
  call    write
  cmp     rax, rdx
  jne     eggog_write
  ret

DEF_EGGOG eggog_imagesize, "image larger than heap"

load_image: ;rdi:path
  ;open file
  mov     rsi, O_RDONLY     ;flags
  call    open_image
  push    rax ;+fd

  ;read register state
  mov     rdi, [rsp]        ;fd
  sub     rsp, HEADER_SIZE
  mov     rsi, rsp          ;buf
  mov     rdx, HEADER_SIZE  ;size
  call    read_image
  pop     B
  pop     F
  pop     L
  pop     S
  pop     E
  pop     C
  pop     D

  ;f reg is also the size of boot image
  shr     F, WORD_SHIFT
  shl     F, WORD_SHIFT
  cmp     F, N
  ja      eggog_imagesize

  ;read heap state
  mov     rdi, [rsp]  ;fd
  mov     rsi, H      ;buf
  mov     rdx, F      ;size
  call    read_image

  ;close file
  pop     rdi ;-fd
  call    close
  ret

DEF_EGGOG eggog_store, "must enable gc to store image"

store_image: ;rdi:path
  ;run gc colector once or twice to defragment heap
  ;and make sure lower heap is active
  test    byte [gc_enabled], 1
  jz      .dont
.gc:
  call    gc
  test    B, B
  jnz     .gc
.dont:

  ;open file
  mov     rsi, O_WRONLY | O_CREAT | O_TRUNC ;flags
  call    open_image
  push    rax ;+fd

  ;write register state
  mov     rdi, [rsp]        ;fd
  push    D
  push    C
  push    E
  push    S
  push    L
  push    F
  push    B
  mov     rsi, rsp          ;buf
  mov     rdx, HEADER_SIZE  ;size
  call    write_image
  add     rsp, HEADER_SIZE

  ;write heap state
  mov     rdi, [rsp]  ;fd
  mov     rsi, H      ;buf
  mov     rdx, F      ;size
  call    write_image

  ;close file
  pop     rdi ;-fd
  call    close
  ret

