%define H rbx ;heap address ptr
%define N r8  ;usable byte size of heap
%define B r9  ;begin active memory area (either 0 or heap_n)
%define F r10 ;next cell returned by _alloc
%define L r11 ;F lock, dont reclaim below this addr
%define S r12 ;stack for intermediate results and retvals
%define E r13 ;environment
%define C r14 ;code
%define D r15 ;dump for storing S E C during calls

section .bss
gc_enabled: resb 1
argc:       resq 1
argv:       resq 1

section .text

str_to_int: ; rsi:str => rbx:int CLOB rcx rax rsi rdx
  ;left to right strlen+1
  xor     rcx, rcx
.strlen:
  inc     rcx
  lodsb
  test    al, al
  jnz     .strlen

  ;right to left multiply digits into rbx
  std
  sub     rsi, 2
  mov     rdx, 1
  xor     rbx, rbx
  jmp     .test ;dec rcx once since strlen+1
.digit:
  xor     rax, rax
  lodsb
  sub     rax, '0'
  cmp     rax, 9
  ja      eggog_usage
  imul    rax, rdx
  add     rbx, rax
  imul    rdx, 10
.test:
  loop    .digit
  cld
  ret

DEF_EGGOG eggog_usage,  "usage: tunavm image heap_bytes gc_0_or_1"
DEF_EGGOG eggog_mmap,   "mmap heap"

init:
  ;dir flag normally clear, any setters must unset when done
  cld

  ;argc >= 4?
  cmp     r14, 4
  jb      eggog_usage

  ;argv[1], image path
  mov     r11, [r15+8]

  ;argv[2], heap size in bytes, mod WORD_SIZE
  mov     rsi, [r15+16]
  call    str_to_int
  shr     rbx, WORD_SHIFT
  shl     rbx, WORD_SHIFT
  mov     rbp, rbx

  ;argv[3], gc enabled 0 | 1
  mov     rsi, [r15+24]
  call    str_to_int
  cmp     rbx, 1
  ja      eggog_usage
  mov     [gc_enabled], bl

  ;store rest of args
  sub     r14, 4
  add     r15, 4*WORD_SIZE
  mov     [argc], r14
  mov     [argv], r15

  ;mmap heap. if gc, mmap(heap*2). otherwise, mmap(heap)
  imul    rbx, rbp
  add     rbx, rbp
  mov     rsi, rbx                          ;size
  ;make sure addresses wont go into type bits in year 3072
  shr     rbx, TYPE_SHIFT
  test    rbx, rbx
  jnz     eggog_mmap
  xor     rdi, rdi                          ;addr
  mov     rdx, PROT_WRITE | PROT_READ       ;prot
  mov     r10, MAP_ANONYMOUS | MAP_PRIVATE  ;flags
  mov     r8,  -1                           ;fd
  xor     r9,  r9                           ;off
  call    mmap
  cmp     rax, 0
  jle     eggog_mmap

  ;set registers
  mov     H, rax ;mmap retval
  mov     N, rbp ;heap_bytes

  ;load image from disk, incl. rest of registers
  mov     rdi, r11
  call    load_image

  ;start machine
  jmp     cycle

