section .text

;fast alloc can be used with up to 8 words (within a op)
%define ALLOC_MAX 8*WORD_SIZE

%macro ALLOC 1 ; bytes | - => rbp:addr
  mov     rbp, F
  add     F, %1
%endmacro

;otherwise, slow alloc must be used
alloc: ;rdx:bytes => rbp:addr CLOB rdx rdi
  mov     rdi, rdx
  mov     rbp, F
  add     rdx, rbp
  mov     F, rdx
  sub     rdx, B
  cmp     rdx, N
  ja      .nomem
  ret
.nomem:
  mov     rdx, rdi
  mov     F, rbp
  test    byte [gc_enabled], 1
  jz      eggog_heap
  call    gc
  mov     rbp, F
  add     rdx, rbp
  mov     F, rdx
  sub     rdx, B
  cmp     rdx, N
  ja      eggog_heap
  ret

DEF_EGGOG eggog_heap, "out of memory"

copy_cell: ; rax:old_addr => rax:new_addr CLOB rcx rdi rsi rbp
  ;if ()/undef/#f/#t, return rax
  cmp     rax, LOWADDR
  jb      .ret

  ;if points to new area, return rax
  mov     rcx, rax
  sub     rcx, N
  lea     rdi, [B-1]
  xor     rcx, rdi
  jns     .ret

  ;if cell moved, return new addr
  lea     rsi, [H+rax]
  mov     rcx, [rsi]
  shr     rcx, TYPE_SHIFT
  cmp     rcx, TYPE_MOVED
  je      .moved

  ;figure cell size
  lea     rdi, [H+F]
  cmp     rcx, TYPE_2WORD
  jbe     .2word
  cmp     rcx, TYPE_VECTOR
  jne     .1word
.vector:
  mov     rcx, [rsi]
  CLEART  rcx
  inc     rcx
  jmp     .copy
.1word:
  mov     rcx, 1
  jmp     .copy
.2word:
  mov     rcx, 2

  ;copy cell words to new area
.copy:
  mov     rbp, F
  lea     F, [F+rcx*WORD_SIZE]
  rep     movsq

  ;mark old cell as moved
  lea     rsi, [H+rax]
  mov     rcx, TYPE_MOVED
  shl     rcx, TYPE_SHIFT
  or      rcx, rbp
  mov     [rsi], rcx

  ;ret new addr
  mov     rax, rbp
.ret:
  ret

.moved:
  mov     rax, [rsi]
  CLEART  rax
  ret

%macro COPY 1
  mov     rax, %1
  call    copy_cell
  mov     %1, rax
%endmacro

gc:
  push    rax
  push    rcx
  push    rdx
  push    rdi
  push    rsi
  push    rbp

  ;swap areas
  test    B, B
  mov     B, 0
  cmovz   B, N
  lea     F, [B+LOWADDR]
  mov     L, F ;L = scan ptr

  ;copy structures from registers
  COPY    S
  COPY    E
  COPY    C
  COPY    D

  ;scan cells in new area for ptrs to olds
  push    S ;need another register
.scan:
  cmp     L, F
  jae     .done

  lea     rdx, [H+L]
  mov     rax, [rdx]
  mov     rcx, rax
  shr     rcx, TYPE_SHIFT
  CLEART  rax

  cmp     rcx, TYPE_PROC ;cons or proc
  jbe     .cons_proc
  cmp     rcx, TYPE_FIXINT
  je      .fixint
  cmp     rcx, TYPE_STRING ;symbol or string
  jbe     .symbol_string
  cmp     rcx, TYPE_VECTOR
  je      .vector

.1word:   ;char/ports
  add     L, WORD_SIZE
  jmp     .scan

.cons_proc:
  add     L, CONS_BYTES ; == PROC_BYTES
  call    copy_cell ;car
  shr     qword [rdx], TYPE_SHIFT
  shl     qword [rdx], TYPE_SHIFT
  or      [rdx], rax
  add     rdx, WORD_SIZE
  COPY    [rdx] ;cdr
  jmp     .scan

.fixint:
  add     L, FIXINT_BYTES
  jmp     .scan

.symbol_string:
  add     L, SYMBOL_BYTES ; == STRING_BYTES
  call    copy_cell ;char vector
  shr     qword [rdx], TYPE_SHIFT
  shl     qword [rdx], TYPE_SHIFT
  or      [rdx], rax
  jmp     .scan

.vector:
  mov     S, rax  ;vec elt count
  lea     L, [L+WORD_SIZE+S*WORD_SIZE]
.copy_elt:
  test    S, S
  jz      .scan
  add     rdx, WORD_SIZE
  COPY    [rdx]
  dec     S
  jmp     .copy_elt

.done:
  pop     S
  mov     L, F ;lock F

  pop     rbp
  pop     rsi
  pop     rdi
  pop     rdx
  pop     rcx
  pop     rax
  ret

