section .rodata
op_tab:
%macro DEF_OP 1 ;label
  dw      _OP%1-_op_offset
%endmacro
%include "defops.s"

section .text
_op_offset:

;-------------------------------------------------------------------------------

;push '()
_OP_NIL:
  ALLOC   CONS_BYTES
  xor     rax, rax
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;load constant from C into S
_OP_LDC:
  ALLOC   CONS_BYTES
  call    pop_c               ;constant
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;get nthcdr of list
nthcdr_next:
  CHECKT  CONS, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rax, [rax]
  dec     rcx
nthcdr: ;rax:l rcx:n => rax:nthcdr(no type check) CLOB rcx rdi
  test    rcx, rcx
  jnz     nthcdr_next
  ret

DEF_EGGOG eggog_type, "type eggog"

;get ptr to word (car of nthcdr of env) containing index of value
;from index into current env
;values are indexed using a cons (x . y) where:
;  x = "relative stack level":
;    0 = local
;    1 = one frame up
;    ...
;  y = value # at that level:
;    0 = (lambda (a b) a)
;    1 = (lambda (a b) b)
;    ...
ref: ;rax:index => rax:car_ptr CLOB rdx rcx rdi rsi
  CHECKT  CONS, rax, rax, rdi
  lea     rdx, [rax+WORD_SIZE]

  ;get x integer value
  mov     rax, [rax]
  CHECKT  FIXINT, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rcx, [rax] ;x

  ;get y integer value
  mov     rax, [rdx]
  CHECKT  FIXINT, rax, rdx, rdi
  add     rdx, WORD_SIZE ;y ptr

  ;get xth value list from env
  mov     rax, E
  call    nthcdr
  CHECKT  CONS, rax, rax, rdi

  ;get yth cdr from value list
  mov     rax, [rax] ;car of env nthcdr
  mov     rcx, [rdx]
  call    nthcdr
  CHECKT  CONS, rax, rax, rdi ;rax now ptr to car of cons
  ret

;load value from extended env into S
_OP_LD:
  ALLOC   CONS_BYTES
  call    pop_c               ;value index
  call    ref
  mov     rax, [rax]          ;unref
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;store value from S into extended env
_OP_ST:
  call    pop_c               ;value index
  call    ref
  CHECKT  CONS, S, rdx, rdi   ;get car of s, leave retval
  mov     rdx, [rdx]          ;rplaca vl with car of s
  mov     [rax], rdx          ;they are both cons so this is easy
  jmp     cycle

;-------------------------------------------------------------------------------

tref:
  call    pop_c               ;cons where cdr = value vector
  CHECKT  CONS, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rax, [rax]          ;value vector
  CHECKT  VECTOR, rax, rdx, rdi
  call    pop_c               ;value index
  CHECKT  FIXINT, rax, rcx, rdi
  ret

;load value from toplvl env into S
_OP_TLD:
  call    tref
  call    vecref
  mov     rax, [rcx]
  cmp     rax, UNDEF
  je      eggog_undef
  ALLOC   CONS_BYTES
  jmp     cycle_push_rax

DEF_EGGOG eggog_undef, "undefined value referenced"

;-------------------------------------------------------------------------------

;store value from S into toplvl env
_OP_TST:
  call    tref
  call    vecref
  CHECKT  CONS, S, rdx, rdi   ;get car of s, leave retval
  mov     rdx, [rdx]
  mov     [rcx], rdx          ;replace value
  jmp     cycle

;-------------------------------------------------------------------------------

;discard car of S, used between statments in block to get rid of return value
_OP_POP:
  call    pop_s
  jmp     cycle

;-------------------------------------------------------------------------------

;pop a thing push its type# as fixint
_OP_TYPE:
  call    pop_s               ;thing to get type of
  ALLOC   FIXINT_BYTES+CONS_BYTES
  lea     rsi, [H+rax]
  lodsq
  shr     rax, TYPE_SHIFT
  mov     rdx, rax
  jmp     cycle_push_fixint

;-------------------------------------------------------------------------------

;vector->symbol
_OP_V2S:
  call    pop_s
  ALLOC   SYMBOL_BYTES+CONS_BYTES
  CHECKT  VECTOR, rax, rdx, rdi
  shl     rax, TYPE_BITS
  or      rax, TYPE_SYMBOL
  jmp     cycle_ror_1

;-------------------------------------------------------------------------------

;vector->string
_OP_V2T:
  call    pop_s
  ALLOC   STRING_BYTES+CONS_BYTES
  CHECKT  VECTOR, rax, rdx, rdi
  shl     rax, TYPE_BITS
  or      rax, TYPE_STRING
cycle_ror_1:
  ror     rax, TYPE_BITS
  lea     rdi, [H+rbp]
  stosq
  mov     rax, rbp
  add     rbp, 1*WORD_SIZE
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;symbol->vector
_OP_S2V:
  call    pop_s
  ALLOC   CONS_BYTES
  CHECKT  SYMBOL, rax, rax, rdi
s2v_t2v:
  mov     rax, [rax]
  CLEART  rax
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;string->vector
_OP_T2V:
  call    pop_s
  ALLOC   CONS_BYTES
  CHECKT  STRING, rax, rax, rdi
  jmp     s2v_t2v

;-------------------------------------------------------------------------------

;char<->fixint conversions must guarante: (for lisp layer stufs to work)
;  1. #\a+1 = #\b, #\b+1 = #\c, ... #\y+1 = #\z
;  2. #\A+1 = #\B, #\B+1 = #\C, ... #\Y+1 = #\Z
;  3. #\0+1 = #\1, #\1+1 = #\2, ... #\8+1 = #\9

;fixint->char
_OP_I2C:
  call    pop_s
  ALLOC   CHAR_BYTES+CONS_BYTES
  CHECKT  FIXINT, rax, rsi, rdi
  add     rsi, WORD_SIZE
  lodsq   ;integer value
  mov     rdx, rax
  shr     rdx, TYPE_SHIFT
  test    rdx, rdx ;(rdx != 0) == (rax < CHAR_MIN || rax > CHAR_MAX)
  jnz     eggog_char
push_char: ;rbp:char_then_cons rax:char_code
  shl     rax, TYPE_BITS
  or      rax, TYPE_CHAR
  jmp     cycle_ror_1

DEF_EGGOG eggog_char, "char range"

;-------------------------------------------------------------------------------

;char->fixint
_OP_C2I:
  call    pop_s
  ALLOC   FIXINT_BYTES+CONS_BYTES
  CHECKT  CHAR, rax, rax, rdi
  mov     rdx, [rax]
  CLEART  rdx
  jmp     cycle_push_fixint

;-------------------------------------------------------------------------------

;pop cons, push its car
_OP_CAR:
  call    pop_s
  ALLOC   CONS_BYTES
  CHECKT  CONS, rax, rax, rdi
  mov     rax, [rax]
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;pop cons, push its cdr
_OP_CDR:
  call    pop_s
  ALLOC   CONS_BYTES
  CHECKT  CONS, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rax, [rax]
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;pop cons and value, rplaca and push cons
_OP_RPLACA:
  call    pop_s               ;cons
  ALLOC   CONS_BYTES
  mov     rcx, rax
  CHECKT  CONS, rax, rsi, rdi
rplaca_rplacd:
  call    pop_s               ;value
  mov     [rsi], rax          ;overwrite car/cdr, TYPE_CONS == 0
  mov     rax, rcx            ;return cons
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;pop cons and value, rplacd and push cons
_OP_RPLACD:
  call    pop_s               ;cons
  ALLOC   CONS_BYTES
  mov     rcx, rax
  CHECKT  CONS, rax, rax, rdi
  lea     rsi, [rax+WORD_SIZE];word1 ptr
  jmp     rplaca_rplacd

;-------------------------------------------------------------------------------

;pop 2 values and push a cons of them
_OP_CONS:
  call    pop_s               ;car
  mov     rcx, rax
  call    pop_s               ;cdr
  ALLOC   CONS_BYTES*2        ;1 for the cons and 1 for pushig to S
  lea     rdi, [H+rbp]
  xchg    rcx, rax
  stosq
  mov     rax, rcx
  jmp     cycle_stosq_2

;-------------------------------------------------------------------------------

;pop vector size and initial element, push allocated vector
_OP_VEC:
  ;get size and chekc
  call    pop_s
  CHECKT  FIXINT, rax, rax, rdi
  lea     rdx, [rax+WORD_SIZE]
  mov     rdx, [rdx]
  mov     rcx, rdx
  shr     rdx, TYPE_SHIFT
  test    rdx, rdx
  jnz     eggog_vecsize

  ;aloc Scons+vector+cells
  lea     rdx, [CONS_BYTES+(1+rcx)*WORD_SIZE]
  call    alloc

  ;initial element into rdx
  call    pop_s
  mov     rdx, rax

  ;set vector cell
  lea     rdi, [H+rbp+CONS_BYTES]
  mov     rax, TYPE_VECTOR
  shl     rax, TYPE_SHIFT
  or      rax, rcx
  stosq

  ;init cells
  mov     rax, rdx
  rep     stosq

  lea     rax, [rbp+CONS_BYTES]
  jmp     cycle_push_rax

DEF_EGGOG eggog_vecsize, "invalid vector size"

;-------------------------------------------------------------------------------

;pop vector, push its size
_OP_VLEN:
  call    pop_s
  ALLOC   FIXINT_BYTES+CONS_BYTES
  CHECKT  VECTOR, rax, rax, rdi
  mov     rdx, [rax]
  CLEART  rdx
  jmp     cycle_push_fixint

;-------------------------------------------------------------------------------

DEF_EGGOG eggog_vecidx, "vector index out of bounds"

check_vecidx: ;rcx:index rdx:vec_ptr => - CLOB rdi
  mov     rdi, [rdx]
  CLEART  rdi
  cmp     rcx, rdi
  jae     eggog_vecidx
  ret

vecref: ;rcx:index_ptr rdx:vec_ptr => rcx:ref CLOB rdi
  ;get index and check bounds
  add     rcx, WORD_SIZE
  mov     rcx, [rcx]
  call    check_vecidx

  ;get word at vec_addr+(1+index)*WORD_SIZE
  inc     rcx
  lea     rcx, [rdx+rcx*WORD_SIZE]
  ret

;pop vector and index, push vector[index]
_OP_VREF:
  call    pop_s ;vec
  CHECKT  VECTOR, rax, rdx, rdi
  call    pop_s ;idx
  CHECKT  FIXINT, rax, rcx, rdi
  ALLOC   CONS_BYTES
  call    vecref
  mov     rax, [rcx]
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;pop vector, index and value, set vector[index] = value, push value
_OP_VSET:
  call    pop_s ;vec
  CHECKT  VECTOR, rax, rdx, rdi
  call    pop_s ;idx
  CHECKT  FIXINT, rax, rcx, rdi
  call    vecref
  CHECKT  CONS, S, rax, rdi
  mov     rax, [rax]
  mov     [rcx], rax
  jmp     cycle

;-------------------------------------------------------------------------------

;pop thing, push #t if '(), otherwise #f
_OP_NULL:
  call    pop_s
  ALLOC   CONS_BYTES
  test    rax, rax
  mov     rax, BOOL_F
  mov     rdx, BOOL_T
  cmovz   rax, rdx
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;pop 2 values and push #t if they have same addr, otherwise #f
_OP_EQ:
  call    pop_s
  mov     rcx, rax
  call    pop_s
  ALLOC   CONS_BYTES
  cmp     rax, rcx
  mov     rax, BOOL_F
  mov     rdx, BOOL_T
  cmove   rax, rdx
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

load_operands: ; - => rdx:x rax:y rbp:cons
  call    pop_s                 ;x
  CHECKT  FIXINT, rax, rcx, rdi
  add     rcx, WORD_SIZE
  call    pop_s                 ;y
  ALLOC   CONS_BYTES
  CHECKT  FIXINT, rax, rsi, rdi
  add     rsi, WORD_SIZE
  lodsq                         ;y
  mov     rdx, [rcx]            ;x
  ret

;fixint =
_OP_EQI:
  call    load_operands
  cmp     rdx, rax
  mov     rax, BOOL_F
  mov     rdx, BOOL_T
  cmove   rax, rdx
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;fixint <
_OP_LT:
  call    load_operands
  cmp     rdx, rax
  mov     rax, BOOL_F
  mov     rdx, BOOL_T
  cmovl   rax, rdx
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;fixint <=
_OP_LE:
  call    load_operands
  cmp     rdx, rax
  mov     rax, BOOL_F
  mov     rdx, BOOL_T
  cmovle  rax, rdx
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

load_operands_fixint: ; - => rdx:x rax:y rbp:fixint_cons
  call    pop_s                 ;x
  CHECKT  FIXINT, rax, rcx, rdi
  add     rcx, WORD_SIZE
  call    pop_s                 ;y
  ALLOC   FIXINT_BYTES+CONS_BYTES
  CHECKT  FIXINT, rax, rsi, rdi
  add     rsi, WORD_SIZE
  lodsq                         ;y
  mov     rdx, [rcx]            ;x
  ret

;fixint +
_OP_ADD:
  call    load_operands_fixint
  add     rdx, rax
  jo      eggog_arithof
  jmp     cycle_push_fixint

;-------------------------------------------------------------------------------

;fixint -
_OP_SUB:
  call    load_operands_fixint
  sub     rdx, rax
  jo      eggog_arithof
  jmp     cycle_push_fixint

DEF_EGGOG eggog_arithof, "arithmetic overflow"

;-------------------------------------------------------------------------------

;fixint *
_OP_MUL:
  call    load_operands_fixint
  imul    rdx, rax
  jo      eggog_arithof
  jmp     cycle_push_fixint

;-------------------------------------------------------------------------------

;fixint / TODO meby this return 2 values
_OP_DIV:
  call    load_operands_fixint
  mov     rcx, rax
  mov     rax, rdx
  test    rcx, rcx
  jz      eggog_divzero
  cmp     rcx, -1
  jne     .ok
  dec     rax
  jo      eggog_arithof
  inc     rax
.ok
  cqo
  idiv    rcx
  mov     rdx, rax
  jmp     cycle_push_fixint

DEF_EGGOG eggog_divzero, "div by zero"

;-------------------------------------------------------------------------------

;pop condition from s and #t/#f branches from C, select
_OP_SEL:
  call    pop_s
  ALLOC   CONS_BYTES          ;cons for pushing C to D
  mov     rcx, rax            ;cond
  call    pop_c
  mov     rsi, rax            ;true branch
  call    pop_c               ;false branch

  ;if cond == #f, overwrite rsi with false branch
  cmp     rcx, BOOL_F
  cmove   rsi, rax

  ;push old C to D and jmp to true/false branch
  lea     rdi, [H+rbp]
  mov     rax, C
  stosq
  mov     rax, D
  stosq
  mov     D, rbp
  mov     C, rsi
  jmp     cycle

;-------------------------------------------------------------------------------

;pop condition from S and #t branch from C, select
;this is necesarry to make _OP_DAP work inside branches,
;cuz nothing must be pushed to D
_OP_TEST:
  call    pop_s
  mov     rcx, rax            ;cond
  call    pop_c               ;true branch

  ;branch if cond != #f
  cmp     rcx, BOOL_F
  cmovne  C, rax
  jmp     cycle

;-------------------------------------------------------------------------------

_OP_JOIN:
  POP_D   C
  jmp     cycle

;-------------------------------------------------------------------------------

;make closure from compiled lambda specified on stack. for eval
_OP_MKF:
  call    pop_s               ;lambda
  mov     rcx, TYPE_PROC
  shl     rcx, TYPE_SHIFT
  or      rcx, rax
  ALLOC   PROC_BYTES+CONS_BYTES
  lea     rdi, [H+rbp]
  mov     [rdi], rcx
  add     rdi, WORD_SIZE
  xor     rax, rax            ;empty env
  jmp     cycle_stosq_2

;-------------------------------------------------------------------------------

;pop lambda from C, push closure of it and current env to S
_OP_LDF:
  call    pop_c               ;lambda
  ALLOC   PROC_BYTES+CONS_BYTES
  lea     rdi, [H+rbp]
  shl     rax, TYPE_BITS
  or      rax, TYPE_PROC
  ror     rax, TYPE_BITS
  stosq
  mov     rax, E              ;+ current env
  jmp     cycle_stosq_2       ;= closure

;-------------------------------------------------------------------------------

;pop closure + value list from S, store S,E,C in D, apply function
_OP_AP:
  call    pop_s                   ;closure
  CHECKT  PROC, rax, rcx, rdi
  call    pop_s                   ;value list
  ALLOC   CONS_BYTES*4            ;conses for pushing S,E,C to D and vl to E
  mov     rdx, rax

  ;push S E C to D
  lea     rdi, [H+rbp]
  mov     rax, S                  ;car of cons0
  stosq
  lea     rax, [rbp+CONS_BYTES*1] ;cdr of cons0 points to cons1
  stosq
  mov     rax, E                  ;car of cons1...
  stosq
  lea     rax, [rbp+CONS_BYTES*2]
  stosq
  mov     rax, C
  stosq
  mov     rax, D                  ;link to old D
  stosq
  mov     D, rbp                  ;put new D

  add     rbp, CONS_BYTES*3
  mov     L, F                    ;lock f
  mov     rax, rdx
  jmp     aply                    ;rest same as _OP_DAP

;-------------------------------------------------------------------------------

;"direct apply", same function as _OP_AP exept no push to D. reqd for tail calls
_OP_DAP:
  call    pop_s                   ;closure
  CHECKT  PROC, rax, rcx, rdi
  call    pop_s                   ;value list
  ALLOC   CONS_BYTES              ;cons for vl to E
  lea     rdi, [H+rbp]
aply:
  xor     S, S                    ;S = EMPTY
  mov     C, [rcx]                ;put closure func in c
  CLEART  C
  stosq                           ;add value list to new env
  add     rcx, WORD_SIZE
  mov     rax, [rcx]              ;link to closure env
  stosq
  mov     E, rbp                  ;put new E
  jmp     cycle

;-------------------------------------------------------------------------------

;restore S,E,C from D and push return value list to S
_OP_RTN:
  mov     rax, S              ;return value list is the state of S
  ;restore registers
  POP_D   S
  POP_D   E
  POP_D   C
  ALLOC   CONS_BYTES
  jmp     cycle_push_rax      ;push return value list

;-------------------------------------------------------------------------------

;checks that car of S is list of length 1, and replaces it with its car
;used to receive a single return value after apply
;also posibly retarded but how else to multiple return values?
_OP_RECV:
  call    pop_s               ;value list
  ALLOC   CONS_BYTES
  lea     rsi, [H+rax]
  lodsq
  mov     rcx, rax            ;car, keep
  shr     rax, TYPE_SHIFT
  test    rax, rax
  jnz     eggog_retval        ;check vl is cons not empty
  lodsq
  test    rax, rax            ;check cdr is EMPTY (vl is length 1)
  jnz     eggog_retval
  mov     rax, rcx
  jmp     cycle_push_rax      ;push the only return value

DEF_EGGOG eggog_retval, "invalid # of return vals"

;-------------------------------------------------------------------------------

;pop arg count from C, check that top lvl env matches exactly
_OP_ARG:
  call    pop_c
  CHECKT  FIXINT, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rcx, [rax]          ;int value
  CHECKT  CONS, E, rax, rdi
  mov     rax, [rax]          ;top lvl env
.loop:                        ;loop through list while decing rcx
  test    rax, rax
  jz      .done
  CHECKT  CONS, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rax, [rax]
  dec     rcx
  jmp     .loop
.done:
  test    rcx, rcx            ;rcx is 0 here if arg# corect
  jz      cycle
  jmp     eggog_argc

DEF_EGGOG eggog_argc, "invalid # of args"

;-------------------------------------------------------------------------------

;pop minimum arg count from C, check that top lvl env has at least that many
;then turn all the rest args into a single list, so it can be loaded as normal
;this is probably retarded, but couldnt thikn of a simpler way
;caller cant know lamda signature at compile time

_OP_ARGV:
  ALLOC   CONS_BYTES          ;cons for making list out of last arg
  call    pop_c               ;min arg count
  CHECKT  FIXINT, rax, rax, rdi
  lea     rcx, [rax+WORD_SIZE]
  mov     rcx, [rcx]          ;int value
  CHECKT  CONS, E, rax, rdi
  mov     rdx, rax            ;need this ptr
  mov     rax, [rax]          ;top lvl env

  ;test for simple (lambda x ...) case
  test    rcx, rcx
  jz      .end2

.loop:
  dec     rcx
  ;check for end of list
  test    rax, rax
  jz      eggog_argc
  ;dont follow cdr on last iteration
  test    rcx, rcx
  jz      .end
  ;(setf rax (cdr rax))
  CHECKT  CONS, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rax, [rax]
  jmp     .loop

  ;finally, (rplacd rax (list (cdr rax)))
.end:
  CHECKT  CONS, rax, rax, rdi
  add     rax, WORD_SIZE
  mov     rdx, rax
  mov     rax, [rax]

  ;use the cons to make a list out of rax and put into [rdx]
.end2:
  lea     rdi, [H+rbp]
  stosq
  xor     rax, rax
  stosq
  mov     [rdx], rbp          ;TYPE_CONS == 0
  jmp     cycle

;-------------------------------------------------------------------------------

;push D register to S, for call/cc
_OP_PUSHD:
  ALLOC   CONS_BYTES
  mov     rax, D
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;pop D register and retval list from S, for call/cc
_OP_POPD:
  call    pop_s
  mov     D, rax
  call    pop_s
  mov     S, rax
  jmp     cycle

;-------------------------------------------------------------------------------

;get stdin port, only used once
_OP_IDEF:
  ALLOC   PORT_BYTES+CONS_BYTES
  mov     rax, STDIN_FILENO
push_iport: ;rbp:port_then_cons rax:fd
  shl     rax, TYPE_BITS
  or      rax, TYPE_IPORT
  jmp     cycle_ror_1

;-------------------------------------------------------------------------------

;get stdout port, only used once
_OP_ODEF:
  ALLOC   PORT_BYTES+CONS_BYTES
  mov     rax, STDOUT_FILENO
push_oport: ;rbp:port_then_cons rax:fd
  shl     rax, TYPE_BITS
  or      rax, TYPE_OPORT
  jmp     cycle_ror_1

;-------------------------------------------------------------------------------

;get stderr port, only used once
_OP_EDEF:
  ALLOC   PORT_BYTES+CONS_BYTES
  mov     rax, STDERR_FILENO
  jmp     push_oport

;-------------------------------------------------------------------------------

%define PATH_BUF 256

conv_path: ;rax:str_ptr rdi:buf => - CLOB not rdx
  ;get char vector
  mov     rax, [rax]
  CLEART  rax
  lea     rsi, [H+rax] ;vec ptr
  lodsq
  mov     rcx, rax
  CLEART  rcx ;vec size

  ;check path len (need 0term)
  cmp     rcx, PATH_BUF
  jae     eggog_path

.loop
  test    rcx, rcx
  jz      .term

  ;get char from next vector elt
  lodsq
  lea     rax, [H+rax]
  mov     rax, [rax]
  CLEART  rax
  cmp     rax, 0xFF
  ja      eggog_char

  ;put in baffer
  stosb
  dec     rcx
  jmp     .loop

.term
  xor     al, al
  stosb
  ret

DEF_EGGOG eggog_path, "path too long"
DEF_EGGOG eggog_open, "open port"

open_port: ; rax:str_ptr rdx:flags => rax:fd
  sub     rsp, PATH_BUF
  mov     rdi, rsp
  call    conv_path
  mov     rdi, rsp        ;path
  mov     rsi, rdx        ;flags
  mov     rdx, FILE_MODE  ;mode
  call    open
  add     rsp, PATH_BUF
  cmp     rax, 0
  jl      eggog_open
  ret

;open input port
_OP_IOPEN:
  call    pop_s
  ALLOC   PORT_BYTES+CONS_BYTES
  CHECKT  STRING, rax, rax, rdi
  mov     rdx, O_RDONLY
  call    open_port
  jmp     push_iport

;-------------------------------------------------------------------------------

;open ouptut port
_OP_OOPEN:
  call    pop_s
  ALLOC   PORT_BYTES+CONS_BYTES
  CHECKT  STRING, rax, rax, rdi
  mov     rdx, O_WRONLY | O_CREAT | O_TRUNC
  call    open_port
  jmp     push_oport

;-------------------------------------------------------------------------------

close_port: ;rax:port_ptr
  mov     rdi, [rax]
  CLEART  rdi
  call    close
  cmp     rax, 0
  jl      eggog_close
  ret

DEF_EGGOG eggog_close, "close port"

;close input port
_OP_ICLOSE:
  ;leave retval
  mov     rax, S
  CHECKT  CONS, rax, rax, rdi
  mov     rax, [rax]
  CHECKT  IPORT, rax, rax, rdi
  call    close_port
  jmp     cycle

;-------------------------------------------------------------------------------

;close output port
_OP_OCLOSE:
  mov     rax, S
  CHECKT  CONS, rax, rax, rdi
  mov     rax, [rax]
  CHECKT  OPORT, rax, rax, rdi
  call    close_port
  jmp     cycle

;-------------------------------------------------------------------------------

;read single char from port
_OP_READC:
  call    pop_s
  CHECKT  IPORT, rax, rax, rdi
  mov     rdi, [rax]
  CLEART  rdi         ;fd
  push    0           ;single char buf, 4 now
  mov     rsi, rsp    ;buf
  mov     rdx, 1      ;size
  call    read
  test    rax, rax
  jz      .eof
  cmp     rax, 1
  jne     eggog_read
  pop     rax 
  ALLOC   CHAR_BYTES+CONS_BYTES
  jmp     push_char
.eof:
  pop     rax
  ALLOC   CONS_BYTES
  mov     rax, BOOL_F
  jmp     cycle_push_rax

DEF_EGGOG eggog_read, "read from port"

;-------------------------------------------------------------------------------

;write single char to port
_OP_WRITEC:
  call    pop_s
  CHECKT  OPORT, rax, rax, rdi
  mov     rdi, [rax]
  CLEART  rdi ;fd
  ;dont pop s to leave retval on stack
  CHECKT  CONS, S, rax, rsi ;char to write
  mov     rax, [rax]
  CHECKT  CHAR, rax, rax, rsi
  mov     rax, [rax]
  CLEART  rax
  cmp     rax, 0xFF
  ja      eggog_char
  push    rax
.write
  mov     rsi, rsp    ;buf
  mov     rdx, 1      ;size
  call    write
  add     rsp, 8
  test    rax, rax
  jz      .write
  cmp     rax, 1
  jne     eggog_write
  jmp     cycle

DEF_EGGOG eggog_write, "write to port"

;-------------------------------------------------------------------------------

;pop item, exit with code 1 if #f, otherwise 0
_OP_STOP:
  call    pop_s
  xor     rdi, rdi
  cmp     rax, BOOL_F
  sete    dil
  call    exit

;-------------------------------------------------------------------------------

;pop path + compiled code, reset machine, replace C with code and dump image
_OP_IMG:
  call    pop_s
  CHECKT  STRING, rax, rax, rdi
  sub     rsp, PATH_BUF
  mov     rdi, rsp
  call    conv_path

  call    pop_s
  xor     S, S
  xor     E, E
  mov     C, rax
  xor     D, D

  mov     rdi, rsp
  call    store_image
  add     rsp, PATH_BUF
  xor     rdi, rdi
  call    exit

;-------------------------------------------------------------------------------

;push speical undefined value to stack
_OP_UNDEF:
  ALLOC   CONS_BYTES
  mov     rax, UNDEF
  jmp     cycle_push_rax

;-------------------------------------------------------------------------------

;strlen incl \0
strlen: ;rsi:str => rdi:len CLOB rsi rax
  xor     rdi, rdi
.loop:
  inc     rdi
  lodsb
  test    al, al
  jz      .ret
  jmp     .loop
.ret:
  ret

;get list of argvs (excl. the mandatory ones)
;"sup" "bich" => #(#\s #\u #\p #\nul #\b #\i #\c #\h #\nul)
;the rest is handled in lispland
_OP_ARGL:
  ;count total chars
  xor     rbp, rbp
  mov     rcx, [argc]
  mov     rdx, [argv]

.loop:
  test    rcx, rcx
  jz      .done
  mov     rsi, [rdx]
  call    strlen
  add     rbp, rdi
  add     rdx, 8
  dec     rcx
  jmp     .loop

.done:
  ;S cons + vector cell + vector elts + chars
  mov     rcx, rbp ;vector elts
  shl     rbp, 1 ;*2 for chars
  lea     rdx, [CONS_BYTES+WORD_SIZE+rbp*8]
  call    alloc

  ;set vector cell
  lea     rdi, [H+rbp+CONS_BYTES]
  mov     rax, TYPE_VECTOR
  shl     rax, TYPE_SHIFT
  or      rax, rcx
  stosq

  ;set elts to point to chars
  lea     rax, [rbp+CONS_BYTES+WORD_SIZE]
  lea     rax, [rax+rcx*WORD_SIZE] ;first char
.loop2:
  test    rcx, rcx
  jz      .done2
  stosq
  add     rax, CHAR_BYTES
  dec     rcx
  jmp     .loop2
.done2:

  ;put chars from argv
  mov     rcx, [argc]
  mov     rdx, [argv]

.loop3:
  test    rcx, rcx
  jz      .done3
  mov     rsi, [rdx]

.loop3b:
  mov     rax, TYPE_CHAR
  shl     rax, TYPE_SHIFT
  lodsb
  stosq
  test    al, al
  jz      .done3b
  jmp     .loop3b
.done3b:

  add     rdx, 8
  dec     rcx
  jmp     .loop3
.done3:

  lea     rax, [rbp+CONS_BYTES]
  jmp     cycle_push_rax

